#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Data {
  [Item("ValueTypeArray", "An abstract base class for representing arrays of value types.")]
  [StorableClass]
  public abstract class ValueTypeArray<T> : Item, IEnumerable<T>, IStringConvertibleArray where T : struct {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Class; }
    }

    [Storable]
    protected T[] array;

    [Storable]
    protected List<string> elementNames;
    public virtual IEnumerable<string> ElementNames {
      get { return this.elementNames; }
      set {
        if (ReadOnly) throw new NotSupportedException("ElementNames cannot be set. ValueTypeArray is read-only.");
        if (value == null || value.Count() == 0)
          elementNames = new List<string>();
        else if (value.Count() != Length)
          throw new ArgumentException("An element name must be specified for each element.");
        else
          elementNames = new List<string>(value);
        OnElementNamesChanged();
      }
    }

    public virtual int Length {
      get { return array.Length; }
      #region Mono Compatibility
      // this setter should be protected, but the Mono compiler couldn't handle it
      set {
        if (ReadOnly) throw new NotSupportedException("Length cannot be set. ValueTypeArray is read-only.");
        if (value != Length) {
          Array.Resize<T>(ref array, value);
          while (elementNames.Count > value)
            elementNames.RemoveAt(elementNames.Count - 1);
          while (elementNames.Count < value)
            elementNames.Add("Element " + elementNames.Count);
          OnElementNamesChanged();
          OnReset();
        }
      }
      #endregion
    }
    public virtual T this[int index] {
      get { return array[index]; }
      set {
        if (ReadOnly) throw new NotSupportedException("Item cannot be set. ValueTypeArray is read-only.");
        if (!value.Equals(array[index])) {
          array[index] = value;
          OnItemsChanged(new List<int>(1) { index });
        }
      }
    }

    public virtual void SetValues(IEnumerable<ArrayValue<T>> arrayValues) {
      if (ReadOnly) throw new NotSupportedException("Item cannot be set. ValueTypeMatrix is read-only.");
      List<int> positions = new List<int>();
      foreach (var arrayValue in arrayValues) {
        if (!arrayValue.Value.Equals(array[arrayValue.Index])) {
          array[arrayValue.Index] = arrayValue.Value;
          positions.Add(arrayValue.Index);
        }
      }
      OnItemsChanged(positions);
    }

    [Storable]
    protected bool readOnly;
    public virtual bool ReadOnly {
      get { return readOnly; }
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (elementNames == null) {
        elementNames = new List<string>(array.Length);
        while (elementNames.Count < array.Length)
          elementNames.Add("Element " + elementNames.Count);
      }
    }

    [StorableConstructor]
    protected ValueTypeArray(bool deserializing) : base(deserializing) { }
    protected ValueTypeArray(ValueTypeArray<T> original, Cloner cloner)
      : base(original, cloner) {
      this.array = (T[])original.array.Clone();
      this.elementNames = new List<string>(original.elementNames);
      this.readOnly = original.readOnly;
    }
    protected ValueTypeArray() {
      array = new T[0];
      elementNames = new List<string>();
      readOnly = false;
    }
    protected ValueTypeArray(int length) {
      array = new T[length];
      elementNames = new List<string>(length);
      readOnly = false;
    }
    protected ValueTypeArray(T[] elements) {
      if (elements == null) throw new ArgumentNullException();
      array = (T[])elements.Clone();
      elementNames = new List<string>(elements.Length);
      readOnly = false;
    }

    public virtual ValueTypeArray<T> AsReadOnly() {
      ValueTypeArray<T> readOnlyValueTypeArray = (ValueTypeArray<T>)this.Clone();
      readOnlyValueTypeArray.readOnly = true;
      return readOnlyValueTypeArray;
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("[");
      if (array.Length > 0) {
        sb.Append(array[0].ToString());
        for (int i = 1; i < array.Length; i++)
          sb.Append(";").Append(array[i].ToString());
      }
      sb.Append("]");
      return sb.ToString();
    }

    public virtual IEnumerator<T> GetEnumerator() {
      return array.Cast<T>().GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }

    public event EventHandler ElementNamesChanged;
    protected virtual void OnElementNamesChanged() {
      EventHandler handler = ElementNamesChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler<ArrayValuesChangedEventArgs> ItemsChanged;
    protected virtual void OnItemsChanged(IEnumerable<int> indices) {
      if (ItemsChanged != null)
        ItemsChanged(this, new ArrayValuesChangedEventArgs(indices));
      OnToStringChanged();
    }
    public event EventHandler Reset;
    protected virtual void OnReset() {
      if (Reset != null)
        Reset(this, EventArgs.Empty);
      OnToStringChanged();
    }

    protected abstract bool Validate(string value, out string errorMessage);
    protected abstract string GetValue(int index);
    protected abstract bool SetValue(string value, int index);
    protected abstract bool SetValues(IEnumerable<ArrayValue<string>> arrayValues);

    #region IStringConvertibleArray Members
    int IStringConvertibleArray.Length {
      get { return Length; }
      set { Length = value; }
    }
    bool IStringConvertibleArray.Validate(string value, out string errorMessage) {
      return Validate(value, out errorMessage);
    }
    string IStringConvertibleArray.GetValue(int index) {
      return GetValue(index);
    }
    bool IStringConvertibleArray.SetValue(string value, int index) {
      return SetValue(value, index);
    }
    bool IStringConvertibleArray.SetValue(ArrayValue<string> arrayValue) {
      return SetValue(arrayValue.Value, arrayValue.Index);
    }
    bool IStringConvertibleArray.SetValues(IEnumerable<ArrayValue<string>> arrayValues) {
      return SetValues(arrayValues);
    }
    #endregion
  }
}
